
--keybinding
renoise.tool():add_keybinding {
  name = "Global:Tools:`VFM` Search and Load VSTi",
  invoke = function()inst_main_search()end  
}

--due to doubling of similar vars/functions in global scope, variables and functions are prefixed with "inst" 
--(vs "fxt" in the effects search)

--readout dimensions
INST_ENTRIES_PER_PAGE = 21
INST_MINIMUM_WIDTH_OF_READOUT = 350
INST_READOUT_HEIGHT = 300

--------------------------------------------------------------------------------
--some basic colors for gui elements
--------------------------------------------------------------------------------
--e.g. For changing inst_vb.views["sample present colour 2"].color when states change
COLOR_GREY = {0x30,0x42,0x42}
COLOR_ORANGE ={0xFF,0x66,0x00}
COLOR_YELLOW = {0xE0,0xE0,0x00}
COLOR_BLUE = {0x50,0x40,0xE0}  
COLOR_RED = {0xEE,0x10,0x10}
COLOR_GREEN = {0x20,0x99,0x20}
COLOR_RED_MILD = {0x90,0x10,0x10}

--Constants holding first index of color-code: Used to identify color in inst_vb.views table i.e.
--if inst_vb.views["button"].color[1] == COLOR_ORANGE_FLAG then --etc
COLOR_GREY_FLAG = 0x30
COLOR_ORANGE_FLAG = 0xFF
COLOR_YELLOW_FLAG = 0xE0
COLOR_BLUE_FLAG = 0x50  
COLOR_RED_FLAG = 0xEE
COLOR_GREEN_FLAG = 0x20
COLOR_RED_MILD_FLAG = COLOR_RED_MILD[1]


--globals
---------
inst_dialog = nil
inst_vb = nil

--popup items
--------------
--holds all available device infos --e.g. renoise.song().tracks[1].available_device_infos 
inst_g_available_device_infos = {} 
--holds all the available infos short_names
inst_popup_names_table = nil
--holds the short names of all the matching plugs from the last search and gets modified to the current search
inst_search_state = {}
inst_search_state.names = {}
inst_search_state.paths = {}
inst_search_state.index = 1
inst_search_state.string = ""
inst_search_state.all_names = {}
inst_search_state.all_paths = {}


---------------------------------------
function inst_initialise_search_state()
---------------------------------------
  
  --1) Empty Search String
  if inst_vb.views["searchfield"].text == "" then
    
    inst_search_state.string = ""
    --default paths table to all paths
    -----------------------------------
    inst_search_state.paths = {"No Match"}
    inst_search_state.names = {"No Match"}
    --add all paths
    for i = 1,#inst_g_available_device_infos do 
      table.insert(inst_search_state.paths,(#inst_search_state.paths+1),inst_g_available_device_infos[i].path)
      table.insert(inst_search_state.names,(#inst_search_state.names+1),inst_g_available_device_infos[i].short_name)
    end
    return
  end
  
  
end


--TODO LOAD BY INDEX NOTE NAME : Need Table updated with indexes

--function to focus the pattern editor 
--(used after a plug is loaded and gui closes)
---------------------------------
function inst_focus_pattern_ed()
---------------------------------
  renoise.app().window.lock_keyboard_focus = not renoise.app().window.lock_keyboard_focus
  renoise.app().window.lock_keyboard_focus = not renoise.app().window.lock_keyboard_focus
end
---

--closes main dialog and removes notifiers
-----------------------------
function inst_closer(dialog)
-----------------------------
  local song = renoise.song()
  local tool = renoise.tool()
  --if dialog exists
  if dialog ~= nil then
    if song.tracks_observable:has_notifier(inst_closer) == true then
      song.tracks_observable:remove_notifier(inst_closer,inst_dialog) --remove notifier
    end
    if tool.app_release_document_observable:has_notifier(inst_closer) == true then
      tool.app_release_document_observable:remove_notifier(inst_closer, inst_dialog) --remove notifier
    end  
  end       
  if dialog and dialog.visible then             
    dialog:close() --close dialog
  end
end
--- 



--load plug from global inst_search_state.paths[] table which is constantly updated
--as the user searches
---------------------------------------------
function inst_load_plug_from_popup_index(idx)
---------------------------------------------
 
  if idx == nil then
    error("Index Not Valid: function: inst_load_plug_from_popup_index()")
    return
  end
  
  local song = renoise.song()
 -- rprint(inst_search_state.paths)
  --load plug
  local plug = song.selected_instrument.plugin_properties:load_plugin(inst_search_state.paths[idx])

  --plug failed to load:
  if plug == false then
    return nil
  end
  
  if options.auto_open_fx.value == true then -- auto_open then
    --open editor
    if song.selected_instrument.plugin_properties.plugin_device.external_editor_available then
      song.selected_instrument.plugin_properties.plugin_device.external_editor_visible = true
    end
  end 
  
  return true
  
end
---

--called for creating/ selecting an empty instrument slot in the list to load vst into --called by New Slot button 
------------------------------------------
function inst_next_empty_instrument_slot()
------------------------------------------
  --song
  local song = renoise.song()
  --loop instruments
  for i = 1,#song.instruments do
    if song.instruments[i].name == "" then
      --select
      song.selected_instrument_index = (i)
      renoise.app():show_status("First Empty Instrument Slot Selected")
      return
    end
  end
  --add as no empty was available
  if song.selected_instrument.name ~= "" then  
    song:insert_instrument_at(#song.instruments + 1)
    song.selected_instrument_index = #song.instruments
    renoise.app():show_status("New Empty Instrument Added ")
  end
end
----

--sets color of New Slot button to red if current inst slot occupied
--grey if not
----------------------------------------
function inst_check_if_inst_slot_clear()
----------------------------------------
  if renoise.song().selected_instrument.name ~= "" then
    return COLOR_RED_MILD
  else
    return COLOR_GREY
  end
end
---


-----------------------------------------
--KEYHANDLER FUNCTION USED FOR SEARCHING
-----------------------------------------

--SEARCH FROM INPUT STRING    
--key handler function() 
-------------------- 
-------------------------------------
function inst_keyhandler_func(dialog, key) 
-------------------------------------
 --rprint(key)
 
  --get user input from "searchfield" 
  local user_input = inst_vb.views["searchfield"].text
  
  --local function containing loops that checks the latest searchfield string against 
  --the available targets in the table. Called on any keyboard input that is not specified otherwise
  ------------------------------
  local function loop_to_match() --called when key == back or key == a character
  ------------------------------
    
    --reset global table
    inst_search_state.names = {}
    --default paths table to all paths
    -----------------------------------
    inst_search_state.paths = {}

  
    --set the user_input string to lower case for comparison
    user_input = string.lower(user_input)
    

    --[Search Loop 1] 
    --Search for popup entries starting with the same letter as the user_input string
    --so when user search begins with "C", all matches with "C" at start are at top of list
    ----------
    local match_at_start = nil
    local added_in_loop_one = {}
    for i = 1,#inst_g_available_device_infos do
      ---make lowercase
      local compared_popup_string = string.lower(inst_g_available_device_infos[i].short_name)
       --use string.find with plain search enabled
      match_at_start = string.find(compared_popup_string,user_input,1,true)
      --if matched from first character, add entry to all matches table
      if match_at_start == 1 then --match found
        --print(inst_popup_names_table[i])
        table.insert(inst_search_state.names,(#inst_search_state.names + 1),inst_g_available_device_infos[i].short_name)
        --MMM
       -- print(inst_g_available_device_infos[i].path) 
        table.insert(inst_search_state.paths,(#inst_search_state.paths + 1),inst_g_available_device_infos[i].path)
        --rprint(inst_search_state.paths)
        table.insert(added_in_loop_one,i)
      end
    end

    --[Search Loop 2] 
    --Search for any sub strings in the popup that match the user_input
    ----------
    for i = 1,#inst_g_available_device_infos do       
      --make lowercase
      local compared_popup_string = string.lower(inst_g_available_device_infos[i].short_name)
      --LOOK FOR MATCH ANYWHERE IN STRING (starting at character 2)
      local found_match = string.find(compared_popup_string,user_input,2,true)--normal
      --check if already added (can occur if letter or short patterns appears twice in the entry)
      local aleady_added = false
      for j = 1,#added_in_loop_one do
        if added_in_loop_one[j] == i then
          aleady_added = true
        end
      end
      
      --this time any match in the string will do
      if (found_match ~= nil) and (aleady_added == false) then --match found
         table.insert(inst_search_state.names,(#inst_search_state.names + 1),inst_g_available_device_infos[i].short_name)
         --MMM 
        -- print(i)
        -- print(inst_popup_names_table[i])
        --  print(inst_g_available_device_infos[i].path)
         table.insert(inst_search_state.paths,(#inst_search_state.paths + 1),inst_g_available_device_infos[i].path)
      end 
    end
    
    --prefix no match when blank search
    local num_finds = #inst_search_state.paths
    if (num_finds + 1) == #inst_search_state.all_paths then --num finds +1 as .all paths includes "No match as entry [1]"
      table.insert(inst_search_state.names,1,"No Match")
      table.insert(inst_search_state.paths,1,"No Match")
    end
    
    --update text readout
    inst_vb.views["readout"].text = inst_tab_to_readout_string(inst_search_state.names)
    --update popup
    inst_vb.views["popup"].items = inst_search_state.names     
  end --end of local function loop_to_match()
  ---
  
  --rprint(inst_search_state)
  
  --[[ --unused   
  -----------------------------------
  local inst_function blank_search_state()   
  -----------------------------------
    --popup
    inst_vb.views["popup"].items = inst_popup_names_table
    inst_vb.views["popup"].value = 1
    --update text readout
    inst_vb.views["readout"].text = ""
  end
  ---
  --]]

  
   -- rprint(key)
  --CURRENT SEARCH TEXT 
  ----------------------
 
  
  --UPDATE SEARCH TEXTFIELD
  --------------------------
  --as function is called before the searchfield is updated we need to keep track
  --by adding the new character here
  if key.character then
    user_input = inst_vb.views["searchfield"].text..key.character --concatenate to build the search string
  end
  --remove end character if backspace
  if key.name == "back" then
    user_input = string.sub(inst_vb.views["searchfield"].text, 1, (#user_input - 1)) --??? potential magical chars problem
  end
  --remove all character if backspace + Ctrl
  if (key.modifiers == "control" and key.name == "back") then
    user_input = "" 
  end
  

  --exit when escape pressed
  ---------------------------
  if (key.modifiers == "" and key.name == "esc") then
    inst_closer(inst_dialog)
  end
  
  --toggle the GUI checkbox
  -------------------------
  if key.name == "f1" or key.name == "f12" then
    inst_vb.views["gui check"].value = (not inst_vb.views["gui check"].value)
    return
  end
  
  --clear the searchfield on ctrl + back or delete
  ----------------------------------------------
  if (key.modifiers == "control" and key.name == "back") or key.name == "del" then
 
      inst_vb.views["searchfield"].text = ""
      --init fn
      inst_initialise_search_state()
      
      inst_vb.views["readout"].text = inst_tab_to_readout_string(inst_search_state.names,1)
      inst_vb.views["popup"].items = inst_search_state.names
      inst_vb.views["popup"].value = 1

    return

      
  --BACK delete last character when back key pressed 
  --------------------------------------------- 
  elseif key.name == "back" then 
    
    local cur_str = inst_vb.views["searchfield"].text 
    --remove last character
    inst_vb.views["searchfield"].text = string.sub(cur_str, 1, (string.len(cur_str)-1))--remove last character from string
    --clear if blank string
    if inst_vb.views["searchfield"].text == "" then
      
      --init fn
      inst_initialise_search_state()
      inst_vb.views["readout"].text = inst_tab_to_readout_string(inst_search_state.names,1)
      inst_vb.views["popup"].items = inst_search_state.names
      inst_vb.views["popup"].value = 1
     
     -- blank_search_state()
      return
    else
      --still characters so compare searchfield to table
      loop_to_match()
      inst_vb.views["popup"].value = 1
    end
    return   
  --UP Arrow chooses previous popup entry   
  ----------------------------------------   
  elseif key.name == "up" then --character input
    if inst_vb.views["popup"].value ~= 1 then
      inst_vb.views["popup"].value = inst_vb.views["popup"].value - 1
    end
    return
  
  --DOWN Arrow  chooses next popup entry
  ----------------------------------------      
  elseif key.name == "down" then --character input
    if inst_vb.views["popup"].value ~= #inst_vb.views["popup"].items then
      inst_vb.views["popup"].value = inst_vb.views["popup"].value + 1 
    end 
    return  
  
  --PRIOR  [PAGE UP]
  --------
  --Go to previous readout page
  elseif key.name == "prior" then
    
    local popup = inst_vb.views["popup"]
    local val = popup.value
    if val >= INST_ENTRIES_PER_PAGE then --there are only 20 entries on page 1 due to paddinf space
      val = val - INST_ENTRIES_PER_PAGE
      --loop back to first on this 'prior page'
      while true do
        if val % INST_ENTRIES_PER_PAGE == 0 then
          break
        end
        val = val - 1
      end
      if val == 0 then --page 1
        val = 1
      end
      popup.value = val
    else --already on page 1 so simply choose first popup value
      popup.value = 1
    end
    return
 
 --NEXT [PAGE DOWN]
 ------
 elseif key.name == "next" then  
  
    local popup = inst_vb.views["popup"]
    local val = popup.value
    --loop + imcfrement val forward
    while val <= #popup.items do
      val = val + 1
      if val % INST_ENTRIES_PER_PAGE == 0 then --will be the first entry of a new page 
         popup.value = val                --(it would be (INST_ENTRIES_PER_PAGE + 1) except we have gui
        break                             --buffer of one empty line on first page)
      end
    end
   return
  
  --HOME
  --------
  elseif key.name == "home" then 
    inst_vb.views["popup"].value = 1
    return
  
  --END
  --------
  elseif key.name == "end" then 
    inst_vb.views["popup"].value = #inst_vb.views["popup"].items
    return
    
   
  --RETURN Execute (load the vst and exit) when return is pressed  
  ---------------------------------------------------------------
  elseif key.name == "return" then
    
    --make sure tables are updated
    local sel_popup_idx = inst_vb.views["popup"].value
    
    inst_search_state.string = inst_vb.views["searchfield"].text
    inst_search_state.index = sel_popup_idx
    
    --if no value chosen then return           
    if (sel_popup_idx == 1) and (inst_vb.views["popup"].items[1] == "No Match") then  
      status("No Match To Load")
      return
    end
    
    local loaded_plug = inst_load_plug_from_popup_index(sel_popup_idx)
    
    --if no plug loaded (string didn't match)
    if loaded_plug == nil then
      status("Plug Failed To Load")
      return
    end

    --update inst_all_matches (used to populate the tool on next opening)
    inst_search_state.names = inst_vb.views["popup"].items
    --focus pattern editor
    inst_focus_pattern_ed()
    --close the GUI
    inst_closer(inst_dialog)
    return


  --CHARACTER
  ------------
  elseif key.character ~= nil then --character input 
    --add character to searchfield
    inst_vb.views["searchfield"].text = inst_vb.views["searchfield"].text..key.character 
    --clear if blank string
    if inst_vb.views["searchfield"].text == "" then
     -- blank_search_state()
      return
    end
    --compare searchfield to table
    loop_to_match() 
    
    inst_vb.views["popup"].value = 1
    return
  end  --elseif any other character
end  

---------------------                     
--- end of keyhandler
---------------------



--used to populate the multiline textfield
-----------------------------------------
function inst_tab_to_readout_string(tab,selected)
-----------------------------------------
  local string = ""
  
  local selected = selected or inst_vb.views["popup"].value 
  local start_line = 1
  
  --SET START LINE
  --1 page has '21' entries (INST_ENTRIES_PER_PAGE), so get start line here:
  if selected == INST_ENTRIES_PER_PAGE then
    start_line = INST_ENTRIES_PER_PAGE --first page has a buffer so special case
  elseif selected > INST_ENTRIES_PER_PAGE then
    --loop back from selected to get readout start line
    for i = selected,1,-1 do
      if i % INST_ENTRIES_PER_PAGE == 0 then
        start_line = i 
        break
      end
    end
  else
    string = "\n" --we are at first page so add buffer line at top
  end

 
  for i = start_line,#tab do
    local entry = inst_search_state.names[i]
    --if is selected in the popup then hilight
    if i == selected  then
      string = string.."  "..entry.."   <----\n"
    else
      string = string.."  "..entry.."\n"
    end
  end
  return string
end --fn
---

  

--Main
---------------------------
function inst_main_search()
---------------------------

  --brings gui to front
  if (inst_dialog and inst_dialog.visible) then -- only allows one dialog instance
    inst_dialog:show() 
    return
  end  
  
  local song = renoise.song() 
  

  -----------------------------------------
  local function inst_get_plugin_type(path)
  -----------------------------------------
    if string.find(path,"VST3/",1,true) then
      return "VST3"
    elseif  string.find(path,"VST/",1,true) then
      return  "VST" 
    elseif  string.find(path,"DSSI",1,true) then
      return  "DSSI"
    elseif  string.find(path,"AU",1,true) then
      return "AU"
    end
    return nil
  end
  ---
  
 
  
  --inst_search_state = {}
  --inst_search_state.names = nil
  --inst_search_state.paths = nil
  --inst_search_state.index = 1
  --inst_search_state.string = ""
  --inst_search_state.all_names = {}
  --inst_search_state.all_paths = {}
  
  --only done on first run
  if #inst_search_state.names == 0 then
    --song list of available plugins/ devices
    -----------------------------------------
    inst_g_available_device_infos = song.instruments[1].plugin_properties.available_plugin_infos
    --sort global table according to short names 
    table.sort(inst_g_available_device_infos , function(a, b) return a.short_name:upper() < b.short_name:upper() end)  
  
    --default search state for names and paths (i.e. all available devices/instruments)
    -----------------------------------
    inst_search_state.names = {"No Match"}
    inst_search_state.paths = {"No Match"}
    --add all paths/short names
    for i = 1,#inst_g_available_device_infos do
      --entries already added
      local num_entries = #inst_search_state.paths
      --short name
      local name = inst_g_available_device_infos[i].short_name
      --path
      local path = inst_g_available_device_infos[i].path
      local plug_type = inst_get_plugin_type(path)
      --if not VST, then modify name in master names table
      if plug_type ~= nil and plug_type == "VST3" then
        name = name.."  ("..plug_type..")"
        inst_g_available_device_infos[i].short_name = name
      end
      --add name
      table.insert(inst_search_state.names,(num_entries+1),name)
      --add path
      table.insert(inst_search_state.paths,(num_entries+1),path)
    end 
    --store all paths/ names
    inst_search_state.all_names = inst_search_state.names
    inst_search_state.all_paths = inst_search_state.paths
  end --first run
  
  
  
  
  
  
 -- rprint(inst_search_state.paths)
   
  --see what the longest name in the search list to set the GUI popup width by
  ---------------------------------
  local function get_longest_name()
  ---------------------------------
    --set name length to 1
    local name_length = 1
    --loop to set to longest string
    for i = 1, #inst_g_available_device_infos do --including ([1] "No Match")
      if inst_g_available_device_infos[i].short_name:len() > name_length then
       name_length = inst_g_available_device_infos[i].short_name:len()
      end
    end
    --multiply the name by 7 (rough average of pixels per character)
    return (name_length * 7)
  end
  ---
 
  local longest_name = get_longest_name()
  longest_name = (longest_name > INST_MINIMUM_WIDTH_OF_READOUT) and longest_name or INST_MINIMUM_WIDTH_OF_READOUT
  
 
 
  --assign global inst_vb
  inst_vb = renoise.ViewBuilder()
  
  ---------------------
  --GUI
  ---------------------
  local inst_dialog_content =
  
    inst_vb:vertical_aligner {
      margin = 8,
      mode = "center",
    
    inst_vb:row {  
      style = "border", --makes a background for the textfield 
   -- inst_vb:text{  
   --  width = 5,
   -- },              
     inst_vb:text{
      style = "strong",
     -- font = "big",
      id = "searchfield",
      width = longest_name,
     -- height = 28,
      text = (inst_search_state.string or "")
     },
    },
                            
    inst_vb:popup {
     id = "popup",
     width = longest_name, 
     value = inst_search_state.index,--default is 1
     items = inst_search_state.all_names,               
     notifier = function() --updates the textfield when an entry is chosen from the pop-up menu
                  
                  if (inst_vb.views["searchfield"].text == "") and
                     (#inst_search_state.names ~= #inst_search_state.all_names) then
                    inst_search_state.names = inst_search_state.all_names
                  end
                  --update readout (multiline searchfield)
                  inst_vb.views["readout"].text = inst_tab_to_readout_string(inst_search_state.names)
                end 
    },
    --spacer text
    inst_vb:horizontal_aligner{
      inst_vb:text{
       text = ""
     },
    },
    
  

    inst_vb:horizontal_aligner{
     -- mode = "center",
     
      inst_vb:text{
        width = 4,--spacer
      },
      
      inst_vb:button{
       id = "slot btn",
       text = "Slot",
       width = 30, 
       height = 20,
       color = inst_check_if_inst_slot_clear(),
       tooltip = "Go To An Empty\nInstrument Slot",
       notifier = function()
                    inst_next_empty_instrument_slot()
                    inst_vb.views["slot btn"].color = COLOR_GREY
                  end,
       
      },
      inst_vb:checkbox{
        id = "slot cb",
        tooltip = "Automatically Choose A New Slot\nWhen Opening Search",
        value = options.auto_inst_new_slot.value,
        notifier = function(value)
                    options.auto_inst_new_slot.value = value
                   end,
      },
      inst_vb:text{
        width = 72,--spacer
      },
     
      inst_vb:button {
        width = 60,
        height = 20,
        text = "Load",
        color = COLOR_GREY,
        notifier = function()

          local sel_popup_idx = inst_vb.views["popup"].value
          --if no value chosen then return           
          if (sel_popup_idx == 1) and (inst_vb.views["popup"].items[1] == "No Match") then  
            status("No Match To Load")
            return
          end
          --set inst_search_state.string so next time opened, the search will be remembered
          inst_search_state.string = inst_vb.views["searchfield"].text
          inst_search_state.index = sel_popup_idx
          --load selected vst in popup
          local loaded_plug = inst_load_plug_from_popup_index(sel_popup_idx)
    
          --if no plug loaded (string didn't match)
          if loaded_plug == nil then
            status("Plug Failed To Load")
            return
          end
      
          --update inst_all_matches (used to populate the tool on next opening)
          inst_search_state.names = inst_vb.views["popup"].items
          --focus pattern editor
          inst_focus_pattern_ed()
          --close the GUI
          inst_closer(inst_dialog)
          return
          
          
        end
      },
      
      inst_vb:text{
       text = "Ed."
      },
      
      --option to auto-open the GUI on Load
      inst_vb:checkbox{
        tooltip = "Auto Opens The Plugs GUI\nEditor On Load",
        value =  options.auto_open_fx.value, 
        notifier = function()  options.auto_open_fx.value = not options.auto_open_fx.value end,
        id = "gui check",
      }
    },--horiz
    inst_vb:text{
     text = "",
    },
    inst_vb:row{
      inst_vb:multiline_textfield{
       id = "readout",
       --align = "top",
       width = longest_name,
       height = INST_READOUT_HEIGHT,
      }
    }
  }
  --end of GUI

  
   
  --SEARCH FROM INPUT STRING    
  --key handler function() was here
  -------------------- 
  
  ------------------------------------------------------------------------------------------
  --Script dialog
  inst_dialog = renoise.app():show_custom_dialog(
      "Search Plugin Instrument", inst_dialog_content,inst_keyhandler_func)
  
  ------------------------------------------------------------------------------------------
  --notifiers
  ------------------------------------------------------------------------------------------

  renoise.tool().app_release_document_observable:add_notifier(inst_closer, inst_dialog)

  ---------------------------------
  --FIRST SETUP/ VALUES FOR READOUT
  ---------------------------------
  --make sure tables are updated
 -- local empty_key = {}
 -- empty_key.character = ""
 -- inst_keyhandler_func(inst_dialog,empty_key)
  ----------------------------
  local function set_to_init()
  ----------------------------
    inst_vb.views["searchfield"].text = inst_search_state.string                    --inst_search_state.index for selected arrow in readout
    inst_vb.views["readout"].text = inst_tab_to_readout_string(inst_search_state.names,inst_search_state.index)
    inst_vb.views["popup"].items = inst_search_state.names
    inst_vb.views["popup"].value = inst_search_state.index
    --inst_vb.views["searchfield"].text = inst_search_state.string
  end
  ---
  
  
  
  --run: set initial search state
  set_to_init()
  
  if inst_vb.views["slot cb"].value == true then
    inst_next_empty_instrument_slot()
    inst_vb.views["slot btn"].color = COLOR_GREY
  end
  

end --end of main
------------------



 
  
  










 
