--keybinding
------------
renoise.tool():add_keybinding {
  name = "Global:Tools:`VFM` Search and Load FX",
  invoke = function()fx_main_search()end
}

--due to doubling of similar vars/functions in global scope, variables and functions are prefixed with "inst" 
--(vs "fxt" in the effects search)

--readout dimensions
FX_ENTRIES_PER_PAGE = 21
FX_MINIMUM_WIDTH_OF_READOUT = 350
FX_READOUT_HEIGHT = 300

--------------------------------------------------------------------------------
--some basic colors for gui elements
--------------------------------------------------------------------------------
--e.g. For changing fx_vb.views["sample present colour 2"].color when states change
COLOR_GREY = {0x30,0x42,0x42}
COLOR_ORANGE ={0xFF,0x66,0x00}
COLOR_YELLOW = {0xE0,0xE0,0x00}
COLOR_BLUE = {0x50,0x40,0xE0}  
COLOR_RED = {0xEE,0x10,0x10}
COLOR_GREEN = {0x20,0x99,0x20}
COLOR_RED_MILD = {0x90,0x10,0x10}

--Constants holding first index of color-code: Used to identify color in fx_vb.views table i.e.
--if fx_vb.views["button"].color[1] == COLOR_ORANGE_FLAG then --etc
COLOR_GREY_FLAG = 0x30
COLOR_ORANGE_FLAG = 0xFF
COLOR_YELLOW_FLAG = 0xE0
COLOR_BLUE_FLAG = 0x50  
COLOR_RED_FLAG = 0xEE
COLOR_GREEN_FLAG = 0x20
COLOR_RED_MILD_FLAG = COLOR_RED_MILD[1]


--globals
---------
fx_dialog = nil
fx_vb = nil

--popup items
--------------
--holds all available device infos --e.g. renoise.song().tracks[1].available_device_infos 
fx_g_available_device_infos = {} 
--holds all the available infos short_names
fx_popup_names_table = nil
--holds the short names of all the matching plugs from the last search and gets modified to the current search
fx_search_state = {}
fx_search_state.names = {}
fx_search_state.paths = {}
fx_search_state.index = 1
fx_search_state.string = ""
fx_search_state.all_names = {}
fx_search_state.all_paths = {}


---------------------------------------
function fx_initialise_search_state()
---------------------------------------
  
  --1) Empty Search String
  if fx_vb.views["searchfield"].text == "" then
    
    fx_search_state.string = ""
    --default paths table to all paths
    -----------------------------------
    fx_search_state.paths = {"No Match"}
    fx_search_state.names = {"No Match"}
    --add all paths
    for i = 1,#fx_g_available_device_infos do 
      table.insert(fx_search_state.paths,(#fx_search_state.paths+1),fx_g_available_device_infos[i].path)
      table.insert(fx_search_state.names,(#fx_search_state.names+1),fx_g_available_device_infos[i].short_name)
    end
    return
  end
  
  
end


--TODO LOAD BY INDEX NOTE NAME : Need Table updated with indexes

--function to focus the pattern editor 
--(used after a plug is loaded and gui closes)
---------------------------------
function fx_focus_pattern_ed()
---------------------------------
  renoise.app().window.lock_keyboard_focus = not renoise.app().window.lock_keyboard_focus
  renoise.app().window.lock_keyboard_focus = not renoise.app().window.lock_keyboard_focus
end
---

--closes main dialog and removes notifiers
-----------------------------
function fx_closer(dialog)
-----------------------------
  local song = renoise.song()
  local tool = renoise.tool()
  --if dialog exists
  if dialog ~= nil then
    if song.tracks_observable:has_notifier(fx_closer) == true then
      song.tracks_observable:remove_notifier(fx_closer,fx_dialog) --remove notifier
    end
    if tool.app_release_document_observable:has_notifier(fx_closer) == true then
      tool.app_release_document_observable:remove_notifier(fx_closer, fx_dialog) --remove notifier
    end  
  end       
  if dialog and dialog.visible then             
    dialog:close() --close dialog
  end
end
--- 


--load plug from global fx_search_state.paths[] table which is constantly updated
--as the user searches
-------------------------------------------
function fx_load_plug_from_popup_index(idx)
-------------------------------------------
 
  if idx == nil then
    error("Index Not Valid: function: fx_load_plug_from_popup_index()")
    return
  end
  
  local song = renoise.song()
  
  --load into sample ed
  if renoise.app().window.active_middle_frame == renoise.ApplicationWindow.MIDDLE_FRAME_INSTRUMENT_SAMPLE_EFFECTS then
    local chain = song.selected_sample_device_chain
    if chain == nil then
      status("No FX Chain Present In Instrument")
      return "No FX Chain Present In Instrument"
    end
    local dev = chain:insert_device_at(fx_search_state.paths[idx], (#chain.devices + 1))
    --select device
    song.selected_sample_device_index = (#chain.devices)
    --open editor if preferenced
    if options.auto_open_fx.value == true then -- auto_open then
      if dev.external_editor_available then
        dev.external_editor_visible = true
      end
    end 
    --return device
    return dev
 
  else--load into track
    local current_track = song.selected_track
    local track_device_total = #song.selected_track.devices
    --add device to track and return it
    local dev = current_track:insert_device_at(fx_search_state.paths[idx], (track_device_total + 1))
    --select device
    song.selected_track_device_index = (track_device_total + 1) 
    --open editor if preferenced
    if options.auto_open_fx.value == true then -- auto_open then
      if dev.external_editor_available then
        dev.external_editor_visible = true
      end
    end 
    --return device
    return dev    
  end
  return nil
end
---

--called for creating/ selecting an empty instrument slot in the list to load vst into --called by New Slot button 
------------------------------------------
function fx_next_empty_instrument_slot()
------------------------------------------
  --song
  local song = renoise.song()
  --loop instruments
  for i = 1,#song.instruments do
    if song.instruments[i].name == "" then
      --select
      song.selected_instrument_index = (i)
      renoise.app():show_status("First Empty Instrument Slot Selected")
      return
    end
  end
  --add as no empty was available
  if song.selected_instrument.name ~= "" then  
    song:insert_instrument_at(#song.instruments + 1)
    song.selected_instrument_index = #song.instruments
    renoise.app():show_status("New Empty Instrument Added ")
  end
end
----

--sets color of New Slot button to red if current inst slot occupied
--grey if not
-----------------------------------
function fx_check_if_fx_slot_clear()
-----------------------------------
  if renoise.song().selected_instrument.name ~= "" then
    return COLOR_RED
  else
    return COLOR_GREY
  end
end
---


-----------------------------------------
--KEYHANDLER FUNCTION USED FOR SEARCHING
-----------------------------------------

--SEARCH FROM INPUT STRING    
--key handler function() 
-------------------- 
-------------------------------------
function fx_keyhandler_func(dialog, key) 
-------------------------------------
 --rprint(key)
 
  --get user input from "searchfield" 
  local user_input = fx_vb.views["searchfield"].text
  
  --local function containing loops that checks the latest searchfield string against 
  --the available targets in the table. Called on any keyboard input that is not specified otherwise
  ------------------------------
  local function loop_to_match() --called when key == back or key == a character
  ------------------------------
    
    --reset global table
    fx_search_state.names = {}
    --default paths table to all paths
    -----------------------------------
    fx_search_state.paths = {}

  
    --set the user_input string to lower case for comparison
    user_input = string.lower(user_input)
    

    --[Search Loop 1] 
    --Search for popup entries starting with the same letter as the user_input string
    --so when user search begins with "C", all matches with "C" at start are at top of list
    ----------
    local match_at_start = nil
    local added_in_loop_one = {}
    for i = 1,#fx_g_available_device_infos do
      ---make lowercase
      local compared_popup_string = string.lower(fx_g_available_device_infos[i].short_name)
       --use string.find with plain search enabled
      match_at_start = string.find(compared_popup_string,user_input,1,true)
      --if matched from first character, add entry to all matches table
      if match_at_start == 1 then --match found
        --print(fx_popup_names_table[i])
        table.insert(fx_search_state.names,(#fx_search_state.names + 1),fx_g_available_device_infos[i].short_name)
        --MMM
       -- print(fx_g_available_device_infos[i].path) 
        table.insert(fx_search_state.paths,(#fx_search_state.paths + 1),fx_g_available_device_infos[i].path)
        --rprint(fx_search_state.paths)
        table.insert(added_in_loop_one,i)
      end
    end

    --[Search Loop 2] 
    --Search for any sub strings in the popup that match the user_input
    ----------
    for i = 1,#fx_g_available_device_infos do       
      --make lowercase
      local compared_popup_string = string.lower(fx_g_available_device_infos[i].short_name)
      --LOOK FOR MATCH ANYWHERE IN STRING (starting at character 2)
      local found_match = string.find(compared_popup_string,user_input,2,true)--normal
      --check if already added (can occur if letter or short patterns appears twice in the entry)
      local aleady_added = false
      for j = 1,#added_in_loop_one do
        if added_in_loop_one[j] == i then
          aleady_added = true
        end
      end
      
      --this time any match in the string will do
      if (found_match ~= nil) and (aleady_added == false) then --match found
         table.insert(fx_search_state.names,(#fx_search_state.names + 1),fx_g_available_device_infos[i].short_name)
         --MMM 
        -- print(i)
        -- print(fx_popup_names_table[i])
        --  print(fx_g_available_device_infos[i].path)
         table.insert(fx_search_state.paths,(#fx_search_state.paths + 1),fx_g_available_device_infos[i].path)
      end 
    end
    
    --prefix no match when blank search
    local num_finds = #fx_search_state.paths
    if (num_finds + 1) == #fx_search_state.all_paths then --num finds +1 as .all paths includes "No match as entry [1]"
      table.insert(fx_search_state.names,1,"No Match")
      table.insert(fx_search_state.paths,1,"No Match")
    end
    
    --update text readout
    fx_vb.views["readout"].text = fx_tab_to_readout_string(fx_search_state.names)
    --update popup
    fx_vb.views["popup"].items = fx_search_state.names     
  end --end of local function loop_to_match()
  ---
  
  --rprint(fx_search_state)
  
  --[[ --unused   
  -----------------------------------
  local fx_function blank_search_state()   
  -----------------------------------
    --popup
    fx_vb.views["popup"].items = fx_popup_names_table
    fx_vb.views["popup"].value = 1
    --update text readout
    fx_vb.views["readout"].text = ""
  end
  ---
  --]]

  
   -- rprint(key)
  --CURRENT SEARCH TEXT 
  ----------------------
 
  
  --UPDATE SEARCH TEXTFIELD
  --------------------------
  --as function is called before the searchfield is updated we need to keep track
  --by adding the new character here
  if key.character then
    user_input = fx_vb.views["searchfield"].text..key.character --concatenate to build the search string
  end
  --remove end character if backspace
  if key.name == "back" then
    user_input = string.sub(fx_vb.views["searchfield"].text, 1, (#user_input - 1)) --??? potential magical chars problem
  end
  --remove all character if backspace + Ctrl
  if (key.modifiers == "control" and key.name == "back") then
    user_input = "" 
  end
  

  --exit when escape pressed
  ---------------------------
  if (key.modifiers == "" and key.name == "esc") then
    fx_closer(fx_dialog)
  end
  
  --toggle the GUI checkbox
  -------------------------
  if key.name == "f1" or key.name == "f12" then
    fx_vb.views["gui check"].value = (not fx_vb.views["gui check"].value)
    return
  end
  
  --clear the searchfield on ctrl + back or delete
  ----------------------------------------------
  if (key.modifiers == "control" and key.name == "back") or key.name == "del" then
 
      fx_vb.views["searchfield"].text = ""
      --init fn
      fx_initialise_search_state()
      
      fx_vb.views["readout"].text = fx_tab_to_readout_string(fx_search_state.names,1)
      fx_vb.views["popup"].items = fx_search_state.names
      fx_vb.views["popup"].value = 1

    return

      
  --BACK delete last character when back key pressed 
  --------------------------------------------- 
  elseif key.name == "back" then 
    
    local cur_str = fx_vb.views["searchfield"].text 
    --remove last character
    fx_vb.views["searchfield"].text = string.sub(cur_str, 1, (string.len(cur_str)-1))--remove last character from string
    --clear if blank string
    if fx_vb.views["searchfield"].text == "" then
      
      --init fn
      fx_initialise_search_state()
      fx_vb.views["readout"].text = fx_tab_to_readout_string(fx_search_state.names,1)
      fx_vb.views["popup"].items = fx_search_state.names
      fx_vb.views["popup"].value = 1
     
     -- blank_search_state()
      return
    else
      --still characters so compare searchfield to table
      loop_to_match()
      fx_vb.views["popup"].value = 1
    end
    return   
  --UP Arrow chooses previous popup entry   
  ----------------------------------------   
  elseif key.name == "up" then --character input
    if fx_vb.views["popup"].value ~= 1 then
      fx_vb.views["popup"].value = fx_vb.views["popup"].value - 1
    end
    return
  
  --DOWN Arrow  chooses next popup entry
  ----------------------------------------      
  elseif key.name == "down" then --character input
    if fx_vb.views["popup"].value ~= #fx_vb.views["popup"].items then
      fx_vb.views["popup"].value = fx_vb.views["popup"].value + 1 
    end 
    return  
  
  --PRIOR  [PAGE UP]
  --------
  --Go to previous readout page
  elseif key.name == "prior" then
    
    local popup = fx_vb.views["popup"]
    local val = popup.value
    if val >= FX_ENTRIES_PER_PAGE then --there are only 20 entries on page 1 due to paddinf space
      val = val - FX_ENTRIES_PER_PAGE
      --loop back to first on this 'prior page'
      while true do
        if val % FX_ENTRIES_PER_PAGE == 0 then
          break
        end
        val = val - 1
      end
      if val == 0 then --page 1
        val = 1
      end
      popup.value = val
    else --already on page 1 so simply choose first popup value
      popup.value = 1
    end
    return
 
 --NEXT [PAGE DOWN]
 ------
 elseif key.name == "next" then  
  
    local popup = fx_vb.views["popup"]
    local val = popup.value
    --loop + imcfrement val forward
    while val <= #popup.items do
      val = val + 1
      if val % FX_ENTRIES_PER_PAGE == 0 then --will be the first entry of a new page 
         popup.value = val                --(it would be (fx_ENTRIES_PER_PAGE + 1) except we have gui
        break                             --buffer of one empty line on first page)
      end
    end
   return
  
  --HOME
  --------
  elseif key.name == "home" then 
    fx_vb.views["popup"].value = 1
    return
  
  --END
  --------
  elseif key.name == "end" then 
    fx_vb.views["popup"].value = #fx_vb.views["popup"].items
    return
    
   
  --RETURN Execute (load the vst and exit) when return is pressed  
  ---------------------------------------------------------------
  elseif key.name == "return" then
    
    local sel_popup_idx = fx_vb.views["popup"].value
    
    fx_search_state.string = fx_vb.views["searchfield"].text
    fx_search_state.index = sel_popup_idx

    --if no value chosen then return           
    if (sel_popup_idx == 1) and (fx_vb.views["popup"].items[1] == "No Match") then  
      status("No Match To Load")
      return
    end
    
    local loaded_plug = fx_load_plug_from_popup_index(sel_popup_idx)
    
    --if no plug loaded (string didn't match)
    if loaded_plug == nil then
      status("Plug Failed To Load")
      return
    end

    --update fx_all_matches (used to populate the tool on next opening)
    fx_search_state.names = fx_vb.views["popup"].items
    --focus pattern editor
    fx_focus_pattern_ed()
    --close the GUI
    fx_closer(fx_dialog)
    return

  --CHARACTER
  ------------
  elseif key.character ~= nil then --character input 
    --add character to searchfield
    fx_vb.views["searchfield"].text = fx_vb.views["searchfield"].text..key.character 
    --clear if blank string
    if fx_vb.views["searchfield"].text == "" then
     -- blank_search_state()
      return
    end
    --compare searchfield to table
    loop_to_match() 
    
    fx_vb.views["popup"].value = 1
    return
  end  --elseif any other character
end  

---------------------                     
--- end of keyhandler
---------------------



--used to populate the multiline textfield
-----------------------------------------
function fx_tab_to_readout_string(tab,selected)
-----------------------------------------
  local string = ""
  
  local selected = selected or fx_vb.views["popup"].value 
  local start_line = 1
  
  --SET START LINE
  --1 page has '21' entries (fx_ENTRIES_PER_PAGE), so get start line here:
  if selected == FX_ENTRIES_PER_PAGE then
    start_line = FX_ENTRIES_PER_PAGE --first page has a buffer so special case
  elseif selected > FX_ENTRIES_PER_PAGE then
    --loop back from selected to get readout start line
    for i = selected,1,-1 do
      if i % FX_ENTRIES_PER_PAGE == 0 then
        start_line = i 
        break
      end
    end
  else
    string = "\n" --we are at first page so add buffer line at top
  end

 
  for i = start_line,#tab do
    local entry = fx_search_state.names[i]
    --if is selected in the popup then hilight
    if i == selected  then
      string = string.."  "..entry.."   <----\n"
    else
      string = string.."  "..entry.."\n"
    end
  end
  return string
end --fn
---

  

--Main
---------------------------
function fx_main_search()
---------------------------

  --brings gui to front
  if (fx_dialog and fx_dialog.visible) then -- only allows one dialog instance
    fx_dialog:show() 
    return
  end  
  
  local song = renoise.song() 
  

  -----------------------------------------
  local function fx_get_plugin_type(path)
  -----------------------------------------
    if string.find(path,"VST3/",1,true) then
      return "VST3"
    elseif  string.find(path,"VST/",1,true) then
      return  "VST" 
    elseif  string.find(path,"DSSI",1,true) then
      return  "DSSI"
    elseif  string.find(path,"AU",1,true) then
      return "AU"
    end
    return nil
  end
  ---
  
 
  
  --fx_search_state = {}
  --fx_search_state.names = nil
  --fx_search_state.paths = nil
  --fx_search_state.index = 1
  --fx_search_state.string = ""
  --fx_search_state.all_names = {}
  --fx_search_state.all_paths = {}
  
  --only done on first run
  if #fx_search_state.names == 0 then
    --song list of available plugins/ devices
    -----------------------------------------
    fx_g_available_device_infos = song.tracks[1].available_device_infos
    --sort global table according to short names 
    table.sort(fx_g_available_device_infos , function(a, b) return a.short_name:upper() < b.short_name:upper() end)  
  
    --default search state for names and paths (i.e. all available devices/instruments)
    -----------------------------------
    fx_search_state.names = {"No Match"}
    fx_search_state.paths = {"No Match"}
    --add all paths/short names
    for i = 1,#fx_g_available_device_infos do
      --entries already added
      local num_entries = #fx_search_state.paths
      --short name
      local name = fx_g_available_device_infos[i].short_name
      --path
      local path = fx_g_available_device_infos[i].path
      local plug_type = fx_get_plugin_type(path)
      --if not VST, then modify name in master names table
      if plug_type ~= nil and plug_type == "VST3" then
        name = name.."  ("..plug_type..")"
        fx_g_available_device_infos[i].short_name = name
      end
      --add name
      table.insert(fx_search_state.names,(num_entries+1),name)
      --add path
      table.insert(fx_search_state.paths,(num_entries+1),path)
    end 
    --store all paths/ names
    fx_search_state.all_names = fx_search_state.names
    fx_search_state.all_paths = fx_search_state.paths
  end --first run
  
  
  
  
  
  
 -- rprint(fx_search_state.paths)
   
  --see what the longest name in the search list to set the GUI popup width by
  ---------------------------------
  local function get_longest_name()
  ---------------------------------
    --set name length to 1
    local name_length = 1
    --loop to set to longest string
    for i = 1, #fx_g_available_device_infos do --including ([1] "No Match")
      if fx_g_available_device_infos[i].short_name:len() > name_length then
       name_length = fx_g_available_device_infos[i].short_name:len()
      end
    end
    --multiply the name by 7 (rough average of pixels per character)
    return (name_length * 7)
  end
  ---
 
  local longest_name = get_longest_name()
  longest_name = (longest_name > FX_MINIMUM_WIDTH_OF_READOUT) and longest_name or FX_MINIMUM_WIDTH_OF_READOUT
  
 
 
  --assign global fx_vb
  fx_vb = renoise.ViewBuilder()
  
  ---------------------
  --GUI
  ---------------------
  local fx_dialog_content =
  
    fx_vb:vertical_aligner {
      margin = 8,
      mode = "center",
    
    fx_vb:row {  
      style = "border", --makes a background for the textfield 
   -- fx_vb:text{  
   --  width = 5,
   -- },              
     fx_vb:text{
      style = "strong",
     -- font = "big",
      id = "searchfield",
      width = longest_name,
     -- height = 28,
      text = (fx_search_state.string or "")
     },
    },
                            
    fx_vb:popup {
     id = "popup",
     width = longest_name, 
     value = fx_search_state.index,--default is 1
     items = fx_search_state.all_names,               
     notifier = function() --updates the textfield when an entry is chosen from the pop-up menu
                  
                  if (fx_vb.views["searchfield"].text == "") and
                     (#fx_search_state.names ~= #fx_search_state.all_names) then
                    fx_search_state.names = fx_search_state.all_names
                  end
                  --update readout (multiline searchfield)
                  fx_vb.views["readout"].text = fx_tab_to_readout_string(fx_search_state.names)
                end 
    },
    --spacer text
    fx_vb:horizontal_aligner{
      fx_vb:text{
       text = ""
     }
    },

    fx_vb:horizontal_aligner{
      mode = "center",
      fx_vb:button {
        width = 60,
        height = 20,
        text = "Load",
        color = COLOR_GREY,
        notifier = function()

          local sel_popup_idx = fx_vb.views["popup"].value
          --if no value chosen then return           
          if (sel_popup_idx == 1) and (fx_vb.views["popup"].items[1] == "No Match") then  
            status("No Match To Load")
            return
          end
          --set fx_search_state.string so next time opened, the search will be remembered
          fx_search_state.string = fx_vb.views["searchfield"].text
          fx_search_state.index = sel_popup_idx
          --load selected vst in popup
          
          --fx_load_plug_from_popup_index(sel_popup_idx) --BUG FIXED by assigning variable here
          local loaded_plug = fx_load_plug_from_popup_index(sel_popup_idx)
          
          --if no plug loaded (string didn't match)
          if loaded_plug == nil then
            status("Plug Failed To Load")
            return
          end  
          --focus pattern editor
          fx_focus_pattern_ed()
          --close the GUI
          fx_closer(fx_dialog)
        end
      },
      
      fx_vb:text{
       text = "Ed."
      },
      
      --option to auto-open the GUI on Load
      fx_vb:checkbox{
        tooltip = "Auto Opens The Plugs GUI\nEditor On Load",
        value =  options.auto_open_fx.value, 
        notifier = function()  options.auto_open_fx.value = not options.auto_open_fx.value end,
        id = "gui check",
      }
    },--horiz
    fx_vb:text{
     text = "",
    },
    fx_vb:row{
      fx_vb:multiline_textfield{
       id = "readout",
       --align = "top",
       width = longest_name,
       height = FX_READOUT_HEIGHT,
      }
    }
  }
  --end of GUI

  
   
  --SEARCH FROM INPUT STRING    
  --key handler function() was here
  -------------------- 
  
  ------------------------------------------------------------------------------------------
  --Script dialog
  fx_dialog = renoise.app():show_custom_dialog(
      "Search FX", fx_dialog_content,fx_keyhandler_func)
  
  ------------------------------------------------------------------------------------------
  --notifiers
  ------------------------------------------------------------------------------------------

  renoise.tool().app_release_document_observable:add_notifier(fx_closer, fx_dialog)

  ---------------------------------
  --FIRST SETUP/ VALUES FOR READOUT
  ---------------------------------
  --make sure tables are updated
 -- local empty_key = {}
 -- empty_key.character = ""
 -- fx_keyhandler_func(fx_dialog,empty_key)
  ----------------------------
  local function set_to_init()
  ----------------------------
    fx_vb.views["searchfield"].text = fx_search_state.string                    --fx_search_state.index for selected arrow in readout
    fx_vb.views["readout"].text = fx_tab_to_readout_string(fx_search_state.names,fx_search_state.index)
    fx_vb.views["popup"].items = fx_search_state.names
    fx_vb.views["popup"].value = fx_search_state.index
    --fx_vb.views["searchfield"].text = fx_search_state.string
  end
  ---
  
  
  
  --run: set initial search state
  set_to_init()
  

end --end of main
------------------



 
  
  










 
