--keybinding
------------
renoise.tool():add_keybinding {
  name = "Global:Tools:`VFM` Search and Load FX",
  invoke = function()fx_main_search()end
}


--due to doubling of similar functions in global scope, variables and functions are prefixed with "fx" 
--(vs "inst" in the instruments search)


ENTRIES_PER_PAGE = 21
MINIMUM_WIDTH_OF_READOUT = 350
READOUT_HEIGHT = 300
 
--globals
---------
fx_dialog = nil
fx_vb = renoise.ViewBuilder()

--popup items
--------------
--holds all available device infos --e.g. renoise.song().tracks[1].available_device_infos 
fx_g_available_device_infos = {} 
--holds all the available infos short_names
fx_popup_names_table = nil
--holds the short names of all the matching plugs from the last search and gets modified to the current search
fx_all_matches = nil
--previous search
-----------------
fx_prev_search = {}
fx_prev_search.index = 1
fx_prev_search.string = ""
fx_prev_search.matches = {}

--function to focus the pattern editor 
--(used after a plug is loaded and gui closes)
------------------------------
function fx_focus_pattern_ed()
------------------------------
  renoise.app().window.lock_keyboard_focus = not renoise.app().window.lock_keyboard_focus
  renoise.app().window.lock_keyboard_focus = not renoise.app().window.lock_keyboard_focus
end
---

--closes main dialog and removes notifiers
--------------------------
function fx_closer(dialog)
--------------------------
  local song = renoise.song()
  local tool = renoise.tool()
  --if dialog exists
  if dialog ~= nil then
    if song.tracks_observable:has_notifier(fx_closer) == true then
      song.tracks_observable:remove_notifier(fx_closer,fx_dialog) --remove notifier
    end
    if tool.app_release_document_observable:has_notifier(fx_closer) == true then
      tool.app_release_document_observable:remove_notifier(fx_closer, fx_dialog) --remove notifier
    end  
  end       
  if dialog and dialog.visible then             
    dialog:close() --close dialog
  end
end
--- 

--gets all the short names for plugin INSTRUMENTS
--------------------------------------
function fx_get_all_plug_short_names()
--------------------------------------
  local song = renoise.song() 
  --Get short names
  local fx_popup_names_table = {}
  --populate table with the plugin short names
  for i = 1,#fx_g_available_device_infos do
    fx_popup_names_table[i] = fx_g_available_device_infos[i].short_name
  end
  --we can now alphabetically sort this table with added function to table.sort()
  --we lower case the short_name keys and compare them (search printout is case in-sensitive which is what we want)
  table.sort(fx_popup_names_table , function(a, b) return a:lower() < b:lower() end)  
  --add a "No Match" field to the beginning
  table.insert(fx_popup_names_table,1,"No Match")
  --return short_names
  return fx_popup_names_table
end


-------------------------------------------
function fx_load_plug_from_short_name(name)
-------------------------------------------

  if name == nil then
    status("Not A Valid Search String")
    return
  end

  --make popup string lower case
  local name = string.lower(name)
  local song = renoise.song()
  local rns_track_device_infos = song.tracks[1].available_device_infos
  --loop all available devices
  for i = 1,#rns_track_device_infos do
    
    --make short_name lower case to comaprer with lowercase name
    local lc_short_name = string.lower(rns_track_device_infos[i].short_name)
    if lc_short_name == name then
      --load into sample ed
      if renoise.app().window.active_middle_frame == renoise.ApplicationWindow.MIDDLE_FRAME_INSTRUMENT_SAMPLE_EFFECTS then
        local chain = song.selected_sample_device_chain
        if chain == nil then
          status("No FX Chain Present In Instrument")
          return nil
        end
        local dev = chain:insert_device_at(rns_track_device_infos[i].path, (#chain.devices + 1))
        --select device
        song.selected_sample_device_index = (#chain.devices)
        --open editor if preferenced
        if options.auto_open_fx.value == true then -- auto_open then
          if dev.external_editor_available then
            dev.external_editor_visible = true
          end
        end 
        --return device
        return dev
      else
        local current_track = song.selected_track
        local track_device_total = #song.selected_track.devices
        --add device to track and return it
        local dev = current_track:insert_device_at(rns_track_device_infos[i].path, (track_device_total + 1))
        --select device
        song.selected_track_device_index = (track_device_total + 1) 
        --open editor if preferenced
        if options.auto_open_fx.value == true then -- auto_open then
          if dev.external_editor_available then
            dev.external_editor_visible = true
          end
        end 
        --return device
        return dev    
      end
    end
  end
  return nil
end




-----------------------------------------
--KEYHANDLER FUNCTION USED FOR SEARCHING
-----------------------------------------

--SEARCH FROM INPUT STRING    
--key handler function() 
-------------------- 
----------------------------------------------
function fx_keyhandler_func(dialog, key) 
----------------------------------------------
 --rprint(key)
 
  --get user input from "searchfield" 
  local user_input = fx_vb.views["searchfield"].text

  --local function containing loops that checks the latest searchfield string against 
  --the available targets in the table. Called on any keyboard input that is not specified otherwise
  ------------------------------
  local function loop_to_match()
  ------------------------------
    
    --reset global table
    fx_all_matches = {}
    --set the user_input string to lower case for comparison

    user_input = string.lower(user_input)

    --[Search Loop 1] 
    --Search for popup entries starting with the same letter as the user_input string
    --so when user search begins with "C", all matches with "C" at start are at top of list
    ----------
    local match_at_start = nil
    local added_in_loop_one = {}
    for i = 2,#fx_popup_names_table do
      ---make lowercase
      local compared_popup_string = string.lower(fx_popup_names_table[i])
       --use string.find with plain search enabled
      match_at_start = string.find(compared_popup_string,user_input,1,true)
      --if matched from first character, add entry to all matches table
      if match_at_start == 1 then --match found
        --print(fx_popup_names_table[i])
        table.insert(fx_all_matches,(#fx_all_matches + 1),fx_popup_names_table[i])
        table.insert(added_in_loop_one,i)
      end
    end

    --[Search Loop 2] 
    --Search for any sub strings in the popup that match the user_input
    ----------
    for i = 2,#fx_popup_names_table do --search from 2 so we don`t include "No Match"        
      --make lowercase
      local compared_popup_string = string.lower(fx_popup_names_table[i])
      --LOOK FOR MATCH ANYWHERE IN STRING (starting at character 2)
      local found_match = string.find(compared_popup_string,user_input,2,true)--normal
      --check if already added (can occur if letter or short patterns appears twice in the entry)
      local aleady_added = false
      for j = 1,#added_in_loop_one do
        if added_in_loop_one[j] == i then
          aleady_added = true
        end
      end
      
      --this time any match in the string will do
      if (found_match ~= nil) and (aleady_added == false) then --match found
         table.insert(fx_all_matches,(#fx_all_matches + 1),fx_popup_names_table[i])
      end 
    end
    
    --update text readout
    fx_vb.views["readout"].text = fx_tab_to_readout_string(fx_all_matches)
    --update popup
    fx_vb.views["popup"].items = fx_all_matches     
  end --end of local function loop_to_match()
  ---
  
  --[[ --unused
  -----------------------------------
  local function blank_search_state()
  -----------------------------------
    --popup
    fx_vb.views["popup"].items = fx_popup_names_table
    fx_vb.views["popup"].value = 1
    --update text readout
    fx_vb.views["readout"].text = ""
  end
  ---
  --]]

  
   -- rprint(key)
  --CURRENT SEARCH TEXT 
  ----------------------
 
  
  --UPDATE SEARCH TEXTFIELD
  --------------------------
  --as function is called before the searchfield is updated we need to keep track
  --by adding the new character here
  if key.character then
    user_input = fx_vb.views["searchfield"].text..key.character --concatenate to build the search string
  end
  --remove end character if backspace
  if key.name == "back" then
    user_input = string.sub(fx_vb.views["searchfield"].text, 1, (#user_input - 1)) --??? potential magical chars problem
  end
  --remove all character if backspace + Ctrl
  if (key.modifiers == "control" and key.name == "back") then
    user_input = "" 
  end
  

  --exit when escape pressed
  ---------------------------
  if (key.modifiers == "" and key.name == "esc") then
    fx_closer(fx_dialog)
  end
  
  --toggle the GUI checkbox
  -------------------------
  if key.name == "f1" or key.name == "f12" then
    fx_vb.views["gui check"].value = (not fx_vb.views["gui check"].value)
    return
  end
  
  --clear the searchfield on ctrl + back or delete
  ----------------------------------------------
  if (key.modifiers == "control" and key.name == "back") or key.name == "del" then
    --clear searchfield to blank
    fx_vb.views["searchfield"].text = ""
    fx_vb.views["readout"].text = fx_tab_to_readout_string(fx_popup_names_table,1)
    fx_vb.views["popup"].items = fx_popup_names_table

    return

      
  --BACK delete last character when back key pressed 
  --------------------------------------------- 
  elseif key.name == "back" then 
    
    local cur_str = fx_vb.views["searchfield"].text 
    --remove last character
    fx_vb.views["searchfield"].text = string.sub(cur_str, 1, (string.len(cur_str)-1))--remove last character from string
    --clear if blank string
    if fx_vb.views["searchfield"].text == "" then
      fx_vb.views["readout"].text = fx_tab_to_readout_string(fx_popup_names_table,1)
      fx_vb.views["popup"].items = fx_popup_names_table
      fx_vb.views["popup"].value = 1
     -- blank_search_state()
      return
    else
      --still characters so compare searchfield to table
      loop_to_match()
      fx_vb.views["popup"].value = 1
    end
    return   
  --UP Arrow chooses previous popup entry   
  ----------------------------------------   
  elseif key.name == "up" then --character input
    if fx_vb.views["popup"].value ~= 1 then
      fx_vb.views["popup"].value = fx_vb.views["popup"].value - 1
    end
    return
  
  --DOWN Arrow  chooses next popup entry
  ----------------------------------------      
  elseif key.name == "down" then --character input
    if fx_vb.views["popup"].value ~= #fx_vb.views["popup"].items then
      fx_vb.views["popup"].value = fx_vb.views["popup"].value + 1 
    end 
    return  
  
  --PRIOR  [PAGE UP]
  --------
  --Go to previous readout page
  elseif key.name == "prior" then
    
    local popup = fx_vb.views["popup"]
    local val = popup.value
    if val >= ENTRIES_PER_PAGE then --there are only 20 entries on page 1 due to paddinf space
      val = val - ENTRIES_PER_PAGE
      --loop back to first on this 'prior page'
      while true do
        if val % ENTRIES_PER_PAGE == 0 then
          break
        end
        val = val - 1
      end
      if val == 0 then --page 1
        val = 1
      end
      popup.value = val
    else --already on page 1 so simply choose first popup value
      popup.value = 1
    end
    return
 
 --NEXT [PAGE DOWN]
 ------
 elseif key.name == "next" then  
  
    local popup = fx_vb.views["popup"]
    local val = popup.value
    --loop + imcfrement val forward
    while val <= #popup.items do
      val = val + 1
      if val % ENTRIES_PER_PAGE == 0 then --will be the first entry of a new page 
         popup.value = val                --(it would be (ENTRIES_PER_PAGE + 1) except we have gui
        break                             --buffer of one empty line on first page)
      end
    end
   return
  
  --HOME
  --------
  elseif key.name == "home" then 
    fx_vb.views["popup"].value = 1
    return
  
  --END
  --------
  elseif key.name == "end" then 
    fx_vb.views["popup"].value = #fx_vb.views["popup"].items
    return
    
   
  --RETURN Execute (load the vst and exit) when return is pressed  
  ---------------------------------------------------------------
  elseif key.name == "return" then
    
    local sel_popup_idx = fx_vb.views["popup"].value
    
    fx_prev_search.string = fx_vb.views["searchfield"].text
    fx_prev_search.index = sel_popup_idx

    --if no value chosen then return           
    if (sel_popup_idx == 1) and (fx_vb.views["popup"].items[1] == "No Match") then  
      status("No Match To Load")
      return
    end
   
    local loaded_plug = fx_load_plug_from_short_name(fx_vb.views["popup"].items[sel_popup_idx])
   
    --if no plug loaded (string didn't match)
    if loaded_plug == nil then
      status("Plug Failed To Load")
      return
    end
    --update fx_all_matches (used to populate the tool on next opening)
    fx_prev_search.matches = fx_vb.views["popup"].items
    --focus pattern editor
    fx_focus_pattern_ed()
    --close the GUI
    fx_closer(fx_dialog)
    return

  --CHARACTER
  ------------
  elseif key.character ~= nil then --character input 
    --add character to searchfield
    fx_vb.views["searchfield"].text = fx_vb.views["searchfield"].text..key.character 
    --clear if blank string
    if fx_vb.views["searchfield"].text == "" then
     -- blank_search_state()
      return
    end
    --compare searchfield to table
    loop_to_match() 
    
    fx_vb.views["popup"].value = 1
    return
  end  --elseif any other character
end  

---------------------                     
--- end of keyhandler
---------------------



--used to populate the multiline textfield
-----------------------------------------
function fx_tab_to_readout_string(tab,selected)
-----------------------------------------
  local string = ""
  local selected = selected or fx_vb.views["popup"].value 
  local start_line = 1
  
  --SET START LINE
  --1 page has '21' entries (ENTRIES_PER_PAGE), so get start line here:
  if selected == ENTRIES_PER_PAGE then
    start_line = ENTRIES_PER_PAGE --first page has a buffer so special case
  elseif selected > ENTRIES_PER_PAGE then
    --loop back from selected to get readout start line
    for i = selected,1,-1 do
      if i % ENTRIES_PER_PAGE == 0 then
        start_line = i 
        break
      end
    end
  else
    string = "\n" --we are at first page so add buffer line at top
  end

  --print(fx_vb.views["popup"].value)
  for i = start_line,#tab do
    --if is selected in the popup then hilight
    if i == selected  then
       string = string.."  "..tab[i].."   <----\n"
      -- string = string..". "..tab[i].."   <----\n"
    else
      --check for doubles which are possible with VST2 and VST3 copies of same plug
      --if tab[i] == tab[i-1] then
        string = string.."  "..tab[i].."\n"
    end
  end
  return string
end --fn
---


  
--------------------------------
function fx_get_plugin_type()
--------------------------------
   --loop to add all plug types to plug_inst_type{}
  local plug_inst_type = {}
  for i = 1,no_of_vsts do
    if  string.find(rns_plugin_infos[i].path , "VST3/") then
      plug_inst_type[i] = "VST3"
    elseif  string.find(rns_plugin_infos[i].path , "VST/") then
      plug_inst_type[i] = "VST" 
    elseif  string.find(rns_plugin_infos[i].path , "DSSI") then
      plug_inst_type[i] = "DSSI"
    elseif  string.find(rns_plugin_infos[i].path ,"AU") then
      plug_inst_type[i] = "AU"
    end
  end
end
---  


--Main
-------------------------
function fx_main_search()
-------------------------

  --brings gui to front
  if (fx_dialog and fx_dialog.visible) then -- only allows one dialog instance
    fx_dialog:show() 
    return
  end  
  
  local song = renoise.song() 
  
  --song list of available plugins/ devices
  -----------------------------------------
  fx_g_available_device_infos = song.tracks[1].available_device_infos
  --This is the list of strings that will appear in the pop-up
  fx_popup_names_table = fx_get_all_plug_short_names()
 
  
  --for first run, make all matches, all of the plugs available
  if fx_all_matches == nil then
    fx_all_matches = fx_popup_names_table
    fx_prev_search.matches = fx_popup_names_table
  end
   
  
  --see what the longest name in the search list to set the GUI popup width by
  ---------------------------------
  local function get_longest_name()
  ---------------------------------
    --set name length to 1
    local name_length = 1
    --loop to set to longest string
    for i = 1, #fx_popup_names_table do --including ([1] "No Match")
      if fx_popup_names_table[i]:len() > name_length then
       name_length = fx_popup_names_table[i]:len()
      end
    end
    --multiply the name by 7 (rough average of pixels per character)
    return (name_length * 7)
  end
  ---
 
  local longest_name = get_longest_name()
  longest_name = (longest_name > MINIMUM_WIDTH_OF_READOUT) and longest_name or MINIMUM_WIDTH_OF_READOUT
  


  --assign global fx_vb
  fx_vb = renoise.ViewBuilder()
  
  ---------------------
  --GUI
  ---------------------
  local fx_dialog_content =
  
    fx_vb:vertical_aligner {
      margin = 8,
      mode = "center",
    
    fx_vb:row {  
      style = "border", --makes a background for the textfield                 
     fx_vb:text{
      style = "strong",
      id = "searchfield",
      width = longest_name,
      text =  ""
     },
    },
                            
    fx_vb:popup {
     id = "popup",
     width = longest_name, 
     value = fx_prev_search.index,--default is 1
     items = fx_popup_names_table,               
     notifier = function() --updates the textfield when an entry is chosen from the pop-up menu
                  
                  if (fx_vb.views["searchfield"].text == "") and
                     (#fx_all_matches ~= #fx_popup_names_table) then
                    fx_all_matches = fx_popup_names_table
                  end
                  --update readout (multiline searchfield)
                  fx_vb.views["readout"].text = fx_tab_to_readout_string(fx_all_matches)
                end 
    },
    --spacer text
    fx_vb:horizontal_aligner{
      fx_vb:text{
       text = ""
     }
    },

    fx_vb:horizontal_aligner{
      mode = "center",
      fx_vb:button {
        width = 60,
        height = 20,
        text = "Load",
        notifier = function()

          local sel_popup_idx = fx_vb.views["popup"].value
          --if no value chosen then return           
          if (sel_popup_idx == 1) and (fx_vb.views["popup"].items[1] == "No Match") then  
            status("No Match To Load")
            return
          end
          --set fx_prev_search.string so next time opened, the search will be remembered
          fx_prev_search.string = fx_vb.views["searchfield"].text
          fx_prev_search.index = sel_popup_idx
          --load selected vst in popup
          fx_load_plug_from_short_name(fx_vb.views["popup"].items[sel_popup_idx])
          --focus pattern editor
          fx_focus_pattern_ed()
          --close the GUI
          fx_closer(fx_dialog)
        end
      },
      
      fx_vb:text{
       text = "Ed."
      },
      
      --option to auto-open the GUI on Load
      fx_vb:checkbox {
        tooltip = "Auto Opens The Plugs GUI\nEditor On Load",
        value =  options.auto_open_fx.value, 
        notifier = function()  options.auto_open_fx.value = not options.auto_open_fx.value end,
        id = "gui check",
      }
    },--horiz
    fx_vb:text{
     text = "",
    },
    fx_vb:row{
      fx_vb:multiline_textfield{
       id = "readout",
       --align = "top",
       width = longest_name,
       height = READOUT_HEIGHT,
      }
    }
  }
  --end of GUI

  
   
  --SEARCH FROM INPUT STRING    
  --key handler function() was here
  -------------------- 
  
  ------------------------------------------------------------------------------------------
  --Script dialog
  fx_dialog = renoise.app():show_custom_dialog(
      "Search FX", fx_dialog_content,fx_keyhandler_func)
  
  ------------------------------------------------------------------------------------------
  --notifiers
  ------------------------------------------------------------------------------------------

  renoise.tool().app_release_document_observable:add_notifier(fx_closer, fx_dialog)

  ---------------------------------
  --FIRST SETUP/ VALUES FOR READOUT
  ---------------------------------
  
  --searchfield
  --readout
  --popup items
  --popup selection
   
  ----------------------------
  local function set_to_init()
  ----------------------------
    fx_vb.views["searchfield"].text = fx_prev_search.string                    --fx_prev_search.index for selected arrow in readout
    fx_vb.views["readout"].text = fx_tab_to_readout_string(fx_prev_search.matches,fx_prev_search.index) 
    fx_vb.views["popup"].items = fx_prev_search.matches
    fx_vb.views["popup"].value = fx_prev_search.index
  end
  ---
  
  --run: set initial search state
  set_to_init()

end --end of main
------------------


